/* This program implements a simple quizcard / flashcard like training
   routine.  It is specifically tailored towards vokabulary training
   (and my main gripe is that I do not know how to handle phonetic
   symbols on a plain DOS-window / XTerm), although it could be used
   for other things.  It should run under both DOS and in a Linux
   Terminal Window (which needs to understand ANSI controll
   sequences).  I am writing this for use on my Sharp PC 3100, a small
   (and rather old) 80C88 based Handhald with a CGA-Display, so you
   might find it somewhat limited by modern Standards...

   This basically implements standard flashcards, but with a twist:
   inside the program each _side_ of a card is represented by a
   different entry.  It is therefore possible for one entry in the
   first language to correspond to more than one entry in the second
   language.

*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>                      /* time() */

/* system-specific stuff.  This should go into a header-file of it's
   own, and possibly even into the makefile ... */
#ifdef unix
#include "conio.h"
#else
#include <conio.h>
#include <string.h>
#endif

#include "quizcard.h"
#include "cconvert.h"
#include "sort.h"
#include "quizfile.h"
#include "browse.h"
#include "test.h"
#include "modeline.h"

/* global variables --- these still need to be made into locals */
unsigned char max_count = 3;           /* number of times to test card */


/* and some more system specific stuff, which, alas, relies on code
   in one of the *.h files */

#ifdef unix
#define CODING_SYSTEM_DEFAULT latin1
#define ESC  27
#define BS  127
#define RET  10
#else
#define CODING_SYSTEM_DEFAULT cp437
#define ESC  27
#define BS    8
#define DEL  83
#endif

/* These functions are missing under Unix */
#ifdef unix
#include <termios.h>
#include <unistd.h>
#include <sys/time.h>
fd_set rfds;
struct timeval tv;
static struct termios stored_settings;
void init_keyboard(void)
{
  struct termios new_settings;
  tcgetattr(0,&stored_settings);
  new_settings = stored_settings;
  /* Disable canonical mode, and set buffer size to 1 byte */
  new_settings.c_lflag &= (~ICANON);
  new_settings.c_cc[VTIME] = 0;
  new_settings.c_cc[VMIN] = 1;
  /* Don't echo characters on the screen */
  new_settings.c_lflag &= (~ECHO);
  tcsetattr(0,TCSANOW,&new_settings);
  return;
}

void restore_keyboard(void)
{
  tcsetattr(0,TCSANOW,&stored_settings);
  return;
}

int kbhit(void)
{
  /* Watch stdin (fd 0) to see when it has input. */
  FD_ZERO(&rfds);
  FD_SET(0, &rfds);
  tv.tv_sec = 0;
  tv.tv_usec = 1;
  return select(1, &rfds, NULL, NULL, &tv);
}
#endif

char readkey(char* chars)
{
  char c;
  char char_entered = 0;
  int i;

  if (chars[0] == 0) {
    char_entered = 1;
  }
#ifdef unix
  /* set terminal into non-canonical mode */
  init_keyboard();
#endif

  do {
    while (!kbhit());
#ifdef unix
    c = getchar();
#else
    c = getch();
#endif
    for (i=0; i<strlen(chars) && !char_entered; i++) {
      if (c == chars[i]) {
        char_entered = 1;
      }
    }
  } while (!char_entered);

#ifdef unix
  /* restore terminal */
  restore_keyboard();
#endif
  
  return c;
}

/* This function allows to add an entry, corresponding to _one_ side
   of a flash-card.  It is then tested

   Each entry will consist of the following fields:

   * word or phrase in language 1
   * wordform: noun/verb/adjective/adverb/preposition/...
   * gender: m/f/n
   * singular or plural
   * a counter

   The calling function needs to provide memory for the entrie's container.

   phrase, translation, and count should be NULL if empty!
*/

int add_entry(entry *the_entry,        /* the new entry */
              char* lang,              /* the name of the language */
              char  *s,		       /* the translated string */
	      int   the_coding_system
              )
{
  char phrase[MAX_PHRASE_LEN+1];       /* pointer to 1st phrase */
  char answer[MAX_PHRASE_LEN+1]="";    /* answer-string */
  char dummy[MAX_PHRASE_LEN+1];        /* string for printing */
  char c;                              /* answer-character */
  char first_pass = TRUE;

  /* the idea here is to have the user enter all Information sequentially,
     then display the information entered and ask whether this information
     is correct.  If it isn't, he can then reenter again, getting prompted
     with the old values were this is reasonable. */

  do {
    /* draw entry-mask */
    clrscr();
    gotoxy(1,1);
    printf("Language %d (%s): %s", ((the_entry->byte1 & LANGUAGE)>>7)+1, lang, reescape(dummy, s));
    if (the_entry->phrase != NULL) {
      gotoxy((80                      /* width of line */
              -strlen(the_entry->phrase) /* length of phrase */
              /*              -strlen(Wordform[(the_entry->byte1 & WORDFORM)>>4]) *//* qualifier */
              -4
              -5)/2,3);
      printf("%s", sprint_phrase(phrase, the_entry));
    } else {
      the_entry->phrase = EMPTY;
    }
    gotoxy(1,4);
    printf("[a] Phrase: %s\n", reescape(dummy, the_entry->phrase));
    gotoxy(1,6);
    printf("[b] Gender:\n");
    printf("    0) noun\n");
    printf("    1) verb\n");
    printf("    2) adjective\n");
    printf("    3) adverb\n");
    printf("    4) preposition\n");
    printf("    7) other\n");
    gotoxy(2,((the_entry->byte1 & WORDFORM)>>4)+7);
    printf("x");
    gotoxy(1,14);
    printf("[c] Gender:\n");
    printf("    m) male\n");
    printf("    f) female\n");
    printf("    n) neuter\n");
    if (((the_entry->byte1 & WORDFORM) == NOUN) ||
        ((the_entry->byte1 & WORDFORM) == ADJECTIV)) {
      gotoxy(2,((the_entry->byte1 & GENDER)>>2)+15);
      printf("x");
    } /* end if NOUN or ADJ. */
    gotoxy(1,19);
    printf("[d] Thingumabob:\n");
    printf("    s) singular\n");
    printf("    p) plural\n");
    if (((the_entry->byte1 & WORDFORM) == NOUN) ||
        ((the_entry->byte1 & WORDFORM) == ADJECTIV)) {
      gotoxy(2,((the_entry->byte1 & CASE)>>1)+20);
      printf("x");
    } /* end if NOUN or ADJ. */
    gotoxy(1,23);
    printf("[q] Quit: ");
    fflush(stdin);
    if (first_pass) {
      /* positioning for hplx */
      gotoxy(1,1);
      c = 'a';
      first_pass = FALSE;
    } else {
      /* positioning for hplx */
      gotoxy(1,4);
      c = readkey("abcdfmnpqs012347ABCDFMNPQS\n\r");
    }
    switch (c) {
    case 'a':
    case 'A':
      gotoxy(13,4);
      /* read answer */
      /* alternatively one can simply press return iff a nonempty default
       * had been provided.  In the latter case the default is kept, while
       * in the former case the memory occupied by the default needs to be
       * freed and new memory needs to be allocated according to the length
       * of the string in answer[], and the answer copied to that string.
       */
      fgets(answer, MAX_PHRASE_LEN, stdin);
      answer[strlen(answer)-1] = 0;
      /* reparse the answer to swap specially encoded characters */
      recode(answer, MAX_PHRASE_LEN+1, the_coding_system);
      if ((the_entry->phrase != NULL) &&
          (strlen(the_entry->phrase) > 0) &&
          (strlen(answer)==0)) {
      } else {
        if (strlen(answer) > 0) {
          if ((the_entry->phrase != NULL) && (the_entry->phrase != EMPTY)) {
            free(the_entry->phrase);
          }
          the_entry->phrase = malloc(strlen(answer)+1);
          if (the_entry->phrase == NULL) {
            perror("Memory for phrase");
            exit(errno);
          }
          strcpy(the_entry->phrase, answer);
        }
      }
      break;
    case 'c':
    case 'C':
      gotoxy(2,((the_entry->byte1 & GENDER)>>2)+15);
      printf(" ");
      fflush(stdin);
      c = readkey("fmnFMN");
    case 'f':
    case 'F':
    case 'm':
    case 'M':
    case 'n':
    case 'N':
      gotoxy(2,((the_entry->byte1 & GENDER)>>2)+15);
      printf(" ");
      switch (c) {
      case 'm':
      case 'M': c = MALE; break;
      case 'n':
      case 'N': c = NEUTER; break;
      case 'f':
      case 'F': c = FEMALE; break;
      }
      the_entry->byte1 = (the_entry->byte1 & -(GENDER+1)) + c;
      gotoxy(2,((the_entry->byte1 & GENDER)>>2)+15);
      printf("x");
      break;
    case 'd':
    case 'D':
      gotoxy(2,((the_entry->byte1 & CASE)>>1)+20);
      printf(" ");
      fflush(stdin);
      c = readkey("spSP");
    case 's':
    case 'S':
    case 'p':
    case 'P':
      gotoxy(2,((the_entry->byte1 & CASE)>>1)+20);
      printf(" ");
      switch (c) {
      case 's':
      case 'S': c = SINGULAR; break;
      case 'p':
      case 'P': c = PLURAL; break;
      }
      the_entry->byte1 = (the_entry->byte1 & -(CASE+1)) + c;
      gotoxy(2,((the_entry->byte1 & CASE)>>1)+20);
      printf("x");
      break;
    case 'b':
    case 'B':
      gotoxy(2,((the_entry->byte1 & WORDFORM)>>4)+7);
      printf(" ");
      fflush(stdin);
      c = readkey("012347");
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '7':
      gotoxy(2,((the_entry->byte1 & WORDFORM)>>4)+7);
      printf(" ");
      the_entry->byte1 = (the_entry->byte1 & -(WORDFORM+1)) + ((c-'0') << 4);
      gotoxy(2,((the_entry->byte1 & WORDFORM)>>4)+7);
      printf("x");
      break;
    } /* end switch */
  } while ((c != 'q') && (c != 'Q') && (c != '\n') && (c != '\r'));

  if (the_entry->phrase == NULL) {
    return 0;
  }
  return strlen(the_entry->phrase);
} /* end add_entry */

/* This function allows to influence all sorts of internal settings.
 * Currently this is mainly the names of the languages and the display's
 * character-set (the character set is internally linked with the languages
 * (so that there are two character-sets), but as long as I only know how to
 * display one set at a time both have of course to be identical...
 *
 * All this information should be kept in a single structure...
 */
int setup(char lang[2][MAX_PHRASE_LEN], word codeset[2], char filename[1024])
{
	char c;
	int  i;
  char answer[MAX_PHRASE_LEN+1]="";

  do {
    clrscr();
    gotoxy(37,5);
    printf("Setup\n");
    gotoxy(1,7);
    printf("1) Language 1: %s\n", lang[0]);
    printf("2) Language 2: %s\n\n", lang[1]);
    printf("   Codeset   : ");
    printf("%s\n", cp_name[codeset[0]]);
    for (i=0; i<NUM_CP; i++) {
      printf("%c) %s\n", 'a'+i, cp_name[i]);
    }
    printf("\n");
    printf("m) Max. count: %d\n\n", max_count);
    printf("f) File-name : %s\n\n", filename);
    printf("q) quit\n");
    c = readkey("12abcdfmqABCDMFQ");
    switch (c) {
    case '1':
      gotoxy(16, 7);
      fgets(answer, MAX_PHRASE_LEN, stdin);
      answer[strlen(answer)-1] = 0;
      /* reparse the answer to swap specially encoded characters */
      recode(answer, MAX_PHRASE_LEN+1, codeset[0]);
      if (strlen(answer) > 0) {
        strcpy(lang[0], answer);
      }
      break;
    case '2':
      gotoxy(16,8);
      fgets(answer, MAX_PHRASE_LEN, stdin);
      answer[strlen(answer)-1] = 0;
      /* reparse the answer to swap specially encoded characters */
      recode(answer, MAX_PHRASE_LEN+1, codeset[1]);
      if (strlen(answer) > 0) {
	strcpy(lang[1], answer);
      }
      break;
    case 'f':
    case 'F':
      clrscr();
      gotoxy(1, 10);
      browse(filename, &filename, 1);
      break;
    case 'a':
    case 'A':
    case 'b':
    case 'B':
    case 'c':
    case 'C':
    case 'd':
    case 'D':
      codeset[0] = c-'a';
      codeset[1] = c-'a';
      break;
    case 'm':
    case 'M':
      gotoxy(16,15);
      fgets(answer, MAX_PHRASE_LEN, stdin);
      answer[strlen(answer)-1] = 0;
      max_count = (char) atoi(answer);
      if (max_count == 0) {
	max_count = 3;
      }
      gotoxy(16,15);
      printf("%d       ", max_count);
      break;
    case 'q':
    case 'Q':
    case '\n':
    case '\r':
      c = 'q';
      break;
    }				       /* end switch */
  } while (c != 'q');
  return 0;
}


int main()
{
  byte  modified  = 0;		       /* flag */
  byte  quit = 0;                      /* flag */
  char  c;                             /* input */
  static entry cards[MAX_NUM_ENTRIES]            = {{0,7<<4,NULL,0,NULL,NULL}};
  static entry *sorted_cards[2][MAX_NUM_ENTRIES] = {{NULL}};
  word  n[3]                                     = {0,0,0};   /* L1, L2, Total */
  word  n_tested[2] = {0,0};	       /* number of tested entries per lang. */
  word  n_pos[2]    = {0,0};	       /* number of cards known */
  word  n_neg[2]    = {0,0};	       /* number of failed cards */
  int   res[2];                        /* flag */
  byte  found[2];                      /* flag */
  int   index[3][2];                   /* position of entries in arrays */
  int   i,j;
  static char  lang[2][MAX_PHRASE_LEN]           = {"Deutsch", "English"};
  int   the_coding_system                        = CODING_SYSTEM_DEFAULT;
  word  codesets[2]                              = {CODING_SYSTEM_DEFAULT, CODING_SYSTEM_DEFAULT};
  byte  l;                             /* current language */
  byte  current_language_test = 0;     /* current language for tests */
  char  filenamestring[1024] = "*.qiz";     /* filename for i/o */
  char  *filename = filenamestring;
  char  substr[MAX_PHRASE_LEN+1];      /* string for search */
  char  dummy[MAX_PHRASE_LEN+1];       /* string for printing */

  /* initialisations */
  clrscr();
  srand((unsigned int)(time(NULL)%RAND_MAX));
  gotoxy(33,10);
  printf("initialising...\n");
  for (i=0; i<MAX_NUM_ENTRIES; i++) {
    cards[i].number = i;
  }

  do {
    clrscr();
    gotoxy(1,1);
    modeline(modified,
	     filename,
	     n,                        /* number of entries */
	     n_tested,                 /* number of tested entries */
	     n_pos,
	     n_neg,
	     lang,
	     the_coding_system,
	     current_language_test);
    gotoxy(30,6);
    printf("Choose next action:\n");
    gotoxy(28,8);
    printf("1) enter new vocabulary\n");
    gotoxy(28,9);
    printf("2) test vocabulary\n");
    gotoxy(28,10);
    printf("3) look up vocabulary\n");
    gotoxy(28,11);
    printf("4) setup\n");
    gotoxy(28,12);
    printf("5) save\n");
    gotoxy(28,13);
    printf("6) load\n");
    gotoxy(28,15);
    printf("q) quit\n");
    gotoxy(28,8);
    fflush(stdin);
    c = readkey("123456qQ\r\n");
    gotoxy(1,16);
    switch (c) {
    case '1':
    case '\r':
    case '\n':
      /* Read in first language */
      cards[n[2]].byte1 = LANGUAGE & LANG0;
      res[0] = add_entry(&(cards[n[2]]), lang[0], "", the_coding_system);
      if (res[0] > 0) {		       /* entry is valid */
        /* look up entry in list */
        index[0][0] = find_entry(&cards[n[2]], sorted_cards[0],
				 n[0], &found[0]);
	if (found[0]) {                /* entry is duplicate entry */
	  index[2][0] = sorted_cards[0][index[0][0]]->number;
          free(cards[n[2]].phrase);
          cards[n[2]].phrase = NULL;
          cards[n[2]].byte1  = 0;
	} else {                       /* entry is new entry */
	  index[2][0] = n[2];
	  n[2]++;
	}
        /*	if (!found[0]) {*/		       /* is a new entry */
        cards[n[2]].byte1 = LANGUAGE & LANG1;
        res[1] = add_entry(&(cards[n[2]]),
			   lang[1],
			   cards[index[2][0]].phrase,
			   the_coding_system);
        if (res[1] > 0) {	       /* entry is valid */
          index[1][1] = find_entry(&cards[n[2]], sorted_cards[1],
                                   n[1], &found[1]);
          if (found[1]) {            /* entry is duplicate entry */
            index[2][1] = sorted_cards[1][index[1][1]]->number;
            free(cards[n[2]].phrase);
            cards[n[2]].phrase = NULL;
            cards[n[2]].byte1  = 0;
          } else {                   /* entry is new entry */
            index[2][1] = n[2];
            n[2]++;
          } /* end entry is new */
        } else {		       /* entry is invalid */
          if (found[0]) {
            /* remove first entry */
            n[2]--;
            free(cards[n[2]].phrase);
            cards[n[2]].phrase = NULL;
            cards[n[2]].byte1  = 0;
          }
        }			       /* end entry 1 is invalid */
        /* }*/			       /* end entry 0 is new */
      }				       /* end entry 0 is valid */
      if ((res[0]>0) && (res[1]>0)) {
	if (!found[0]) {
	  /* add entry */
	  insert_entry(&cards[index[2][0]],
		       sorted_cards[0], n[0], index[0][0]);
	  n[0]++;
	}
	if (!found[1]) {
	  /* add entry */
	  insert_entry(&cards[index[2][1]],
		       sorted_cards[1], n[1], index[1][1]);
	  n[1]++;
	}
	/* crosslink entries */
	cards[index[2][0]].num_translations++;
	cards[index[2][0]].translations =
	  realloc(cards[index[2][0]].translations,
		  cards[index[2][0]].num_translations
		  *sizeof(cards[index[2][0]].translations[0]));
	cards[index[2][0]].count =
	  realloc(cards[index[2][0]].count,
		  cards[index[2][0]].num_translations
		  *sizeof(cards[index[2][0]].count[0]));
	cards[index[2][0]].translations[cards[index[2][0]].num_translations-1] = index[2][1];
	cards[index[2][0]].count[cards[index[2][0]].num_translations-1] = 0;

	cards[index[2][1]].num_translations++;
	cards[index[2][1]].translations =
	  realloc(cards[index[2][1]].translations,
		  cards[index[2][1]].num_translations
		  *sizeof(cards[index[2][1]].translations[0]));
	cards[index[2][1]].count =
	  realloc(cards[index[2][1]].count,
		  cards[index[2][1]].num_translations
		  *sizeof(cards[index[2][1]].count[0]));
	cards[index[2][1]].translations[cards[index[2][1]].num_translations-1] = index[2][0];
	cards[index[2][1]].count[cards[index[2][1]].num_translations-1] = 0;
	modified = 1;
      }				       /* end if both entries valid */
      break;
    case '2':
      /* test vocabulary */
      test(cards, n, n_tested, n_pos, n_neg, lang, the_coding_system,
	   &current_language_test);
      modified = 1;
      break;
    case '3':
      clrscr();
      /* select language */
      gotoxy(20,5);
      printf("1) look up a phrase in %s", lang[0]);
      gotoxy(20,6);
      printf("2) look up a phrase in %s", lang[1]);
      gotoxy(20,7);
      printf("3) look up a substring in %s", lang[0]);
      gotoxy(20,8);
      printf("4) look up a substring in %s", lang[1]);
      gotoxy(20,9);
      printf("q) quit");
      gotoxy(20,9);
      fflush(stdout);
      fflush(stdin);
      c = readkey("12345qQ\r\n");
      switch (c) {
      case '1':
      case '2':
      case '3':
      case '4':
        j = c-'1';
        i = 0;
        substr[0] = 0;
        clrscr();
        fflush(stdout);
        c=0;
        while ((i<1023) && (c != ESC)) {
          clrscr();
          gotoxy(1,1);
          recode(substr, MAX_NUM_ENTRIES, the_coding_system);
          i = strlen(substr);
          if (c != '\\') {
            reescape(dummy, substr);
          } else {
            sprintf(dummy, "%s", substr);
          }
          printf("%3d: %s\n", c, dummy);
          switch (j) {
          case 0:
          case 1:
            show_startstr(cards,
                          sorted_cards[j],
                          n[j],
                          substr,
                          8);
            break;
          case 2:
          case 3:
            show_substr(cards,
                        sorted_cards[j-2],
                        n[j-2],
                        substr,
                        8);
            break;
          }
          gotoxy(6+strlen(dummy),1);
          fflush(stdout);
          fflush(stdin);
          c = readkey("");
          substr[i++] = c;
          if (c == BS) {
            /* BS on Zaurus */
            i-=2;
          }
          if (i >= 0) {
            substr[i] = 0;
          } else {
            i = 0;
            c = ESC;              /* simulate ESC */
          }
        } /* end while */
        break;
      case '5':
      case 'q':
      case 'Q':
      case '\r':
      case '\n':
      default:
        break;
      } /* end switch */
      break;
    case '4':
      /* setup */
      setup(lang, codesets, filename);
      the_coding_system = codesets[0];
      modified = 1;
      break;
    case '5':
      /* save */
      browse(filename, &filename, 1);
      if (filename[0] == 0) {
        break;
      }
      write_quiz_file(filename, lang, codesets, n[2], current_language_test, cards);
      modified = 0;
      break;
    case '6':
      /* load */
      /* do I have to save some modifications first? */
      if (modified) {
	printf("\nCards have been modified, you should save first - really load? [N/y]\n");
        c = readkey("yYnN\r\n");
	if ( c != 'y' && c != 'Y' ) {
	  break;
	}
      } /* end if modified */
      /* Do I need to clear cards first, in case we are loading after some
         cards had already been entered? */
      browse(filename, &filename, 0);
      if (filename[0] == 0) {
        break;
      }
      n_pos[0] = 0;
      n_pos[1] = 0;
      n_neg[0] = 0;
      n_neg[1] = 0;
      read_quiz_file(filename, lang, codesets, the_coding_system, &(n[2]), &current_language_test, cards);
      printf("\nPress 'q' when ready\n");
      readkey("qQ\r\n");
      /* enter cards into sorted lists */
      fprintf(stderr, "inserting %d cards:\n", n[2]);
      /* reset crosslinks */
      for (i=0;i<n[0]; i++) {
	sorted_cards[0][i] = NULL;
      }
      n[0]=0;
      for (i=0;i<n[1]; i++) {
	sorted_cards[1][i] = NULL;
      }
      n[1]=0;
      n_tested[0] = 0;
      n_tested[1] = 0;
      for (i=0; i<n[2]; i++) {
	l = (cards[i].byte1 & LANGUAGE) >> 7;
	if (cards[i].byte1 & TESTED) {
	  /* Cards can be marked as tested for one of two reasons:
	   * 1) The file has been saved _during_ a test, in which case
	   *    some cards have indeed been tested already.
	   * 2) A card has passed it's final test (3 times) and need not 
	   *    be tested any further.
	   */
	  n_tested[l]++;
	}
	/* fprintf(stderr, "Entry %d: %s\n", i, lang[l]); */
	index[2][l] = find_entry(&(cards[i]), sorted_cards[l],
				 n[l], &found[l]);
	insert_entry(&(cards[i]), sorted_cards[l], n[l], index[2][l]);
	n[l]++;
      }				       /* end for i */
      printf("%d cards in %s, %d/%d = %4.1f%% tested\n", n[0], lang[0], n_tested[0], n[0], n_tested[0]*100.0/n[0]);
      printf("%d cards in %s, %d/%d = %4.1f%% tested\n", n[1], lang[1], n_tested[1], n[1], n_tested[1]*100.0/n[1]);
      printf("\nPress 'q' when ready\n");
      readkey("qQ\r\n");
      modified = 0;
      break;
    case 'q':
    case 'Q':
      /* do I have to save some modifications first? */
      if (modified) {
	printf("\nCards have been modified, you should save first - really quit? [N/y]\n");
        c = readkey("yYnN\r\n");
	if ( c != 'y' && c != 'Y' ) {
	  break;
	}
      } /* end if modified */
      quit = 1;
      break;
    }
  } while (!quit);
  return 0;
}
